﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2014 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Drawing;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.EvolutionTracking {
  [Item("DirectedGraph", "Generic class representing a directed graph with custom vertices and content")]
  [StorableClass]
  public class DirectedGraph<TVertex> : Item, IDirectedGraph<TVertex> where TVertex : class,IVertex {
    [Storable]
    protected readonly List<TVertex> nodes; // for performance reasons, maybe this should be a linked list (fast remove)
    public List<TVertex> Nodes {
      get { return nodes; }
    }
    public DirectedGraph() {
      nodes = new List<TVertex>();
    }
    [StorableConstructor]
    protected DirectedGraph(bool serializing)
      : base(serializing) {
    }
    protected DirectedGraph(DirectedGraph<TVertex> original, Cloner cloner)
      : base(original, cloner) {
      nodes = new List<TVertex>(original.Nodes);
    }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new DirectedGraph<TVertex>(this, cloner);
    }
    public bool Contains(TVertex t) {
      return nodes.Contains(t);
    }

    public virtual bool Any(Func<TVertex, bool> predicate) {
      return nodes.Any(predicate);
    }
    public virtual bool IsEmpty {
      get { return !nodes.Any(); }
    }

    public virtual void Clear() {
      nodes.Clear();
    }

    public virtual void AddVertex(TVertex vertex) {
      Nodes.Add(vertex);
    }

    public virtual void RemoveVertex(TVertex vertex) {
      nodes.Remove(vertex);
    }
    public override Image ItemImage {
      get { return Common.Resources.VSImageLibrary.Graph; }
    }
  }

  [Item("Directed graph", "Generic directed graph base class.")]
  [StorableClass]
  public class DirectedGraph<TVertex, TContent> : DirectedGraph<TVertex>, IDirectedGraph<TVertex, TContent>
    where TVertex : class,IVertex<TContent>
    where TContent : class,IItem {

    [StorableConstructor]
    protected DirectedGraph(bool deserializing) : base(deserializing) { }

    [Storable]
    private readonly Dictionary<TContent, List<TVertex>> contentMap; // reverse mapping of content to the actual vertices
    public DirectedGraph() {
      contentMap = new Dictionary<TContent, List<TVertex>>();
    }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new DirectedGraph<TVertex, TContent>(this, cloner);
    }
    protected DirectedGraph(DirectedGraph<TVertex, TContent> original, Cloner cloner)
      : base(original, cloner) {
      contentMap = new Dictionary<TContent, List<TVertex>>(original.contentMap);
    }
    public bool Contains(TContent content) {
      return contentMap.ContainsKey(content);
    }
    public List<TVertex> this[TContent key] {
      get {
        List<TVertex> result = null;
        contentMap.TryGetValue(key, out result);
        return result;
      }
      //      set {
      //        contentMap[key] = value;
      //      }
    }
    public override void Clear() {
      contentMap.Clear();
      base.Clear();
    }
    public override void AddVertex(TVertex vertex) {
      if (contentMap.ContainsKey(vertex.Content)) {
        contentMap[vertex.Content].Add(vertex);
      } else {
        contentMap[vertex.Content] = new List<TVertex> { vertex };
      }
      base.AddVertex(vertex);
    }

    public override void RemoveVertex(TVertex vertex) {
      if (contentMap.ContainsKey(vertex.Content)) {
        contentMap[vertex.Content].Remove(vertex);
      }
      base.RemoveVertex(vertex);
    }
  }
}
