﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2014 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.EvolutionTracking {
  [StorableClass]
  [Item("GenealogyGraph", "A class representing a genealogy graph")]
  public class GenealogyGraph : GenealogyGraph<IGenealogyGraphNode> {
  }

  [StorableClass]
  [Item("GenealogyGraph", "A class representing a genealogy graph (of a population of individuals) without storing any content for vertices")]
  public class GenealogyGraph<TVertex> : DirectedGraph<TVertex>, IGenealogyGraph<TVertex> where TVertex : class,IGenealogyGraphNode {
    [Storable]
    private Dictionary<double, LinkedList<TVertex>> ranks; // use a linked list for fast insertion/removal

    public Dictionary<double, LinkedList<TVertex>> Ranks {
      get { return ranks; }
      set { ranks = value; }
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new GenealogyGraph<TVertex>(this, cloner);
    }
    protected GenealogyGraph(GenealogyGraph<TVertex> original, Cloner cloner)
      : base(original, cloner) {
    }
    [StorableConstructor]
    protected GenealogyGraph(bool deserializing) : base(deserializing) { }

    public GenealogyGraph() {
      Ranks = new Dictionary<double, LinkedList<TVertex>>();
    }
    public override void AddVertex(TVertex vertex) {
      if (!Ranks.ContainsKey(vertex.Rank)) {
        Ranks[vertex.Rank] = new LinkedList<TVertex>();
      }
      Ranks[vertex.Rank].AddLast(vertex);
      base.AddVertex(vertex);
    }
    public override void RemoveVertex(TVertex vertex) {
      if (Ranks.ContainsKey(vertex.Rank)) {
        Ranks[vertex.Rank].Remove(vertex);
      }
      base.RemoveVertex(vertex);
    }
    // updated event
    public event EventHandler GraphUpdated;
    private void OnGraphUpdated(object sender, EventArgs args) {
      var updated = GraphUpdated;
      if (updated != null) updated(sender, args);
    }
  }

  [StorableClass]
  [Item("GenealogyGraph",
    "A class representing a genealogy graph (of a population of individuals) in which each vertex stores a content.")]
  public class GenealogyGraph<TVertex, TContent> : DirectedGraph<TVertex, TContent>, IGenealogyGraph<TVertex, TContent>
    where TVertex : class, IGenealogyGraphNode<TContent>
    where TContent : class, IItem {
    [Storable]
    private Dictionary<double, LinkedList<TVertex>> ranks;
    public Dictionary<double, LinkedList<TVertex>> Ranks {
      get { return ranks; }
      set { ranks = value; }
    }
    protected GenealogyGraph(GenealogyGraph<TVertex, TContent> original, Cloner cloner)
      : base(original, cloner) {
    }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new GenealogyGraph<TVertex, TContent>(this, cloner);
    }
    public GenealogyGraph() {
      Ranks = new Dictionary<double, LinkedList<TVertex>>();
    }
    public override void AddVertex(TVertex vertex) {
      if (!Ranks.ContainsKey(vertex.Rank)) {
        Ranks[vertex.Rank] = new LinkedList<TVertex>();
      }
      Ranks[vertex.Rank].AddLast(vertex);
      base.AddVertex(vertex);
    }
    public override void RemoveVertex(TVertex vertex) {
      if (Ranks.ContainsKey(vertex.Rank)) {
        Ranks[vertex.Rank].Remove(vertex);
      }
      base.RemoveVertex(vertex);
    }
    // updated event
    public event EventHandler GraphUpdated;
    private void OnGraphUpdated(object sender, EventArgs args) {
      var updated = GraphUpdated;
      if (updated != null) updated(sender, args);
    }
  }
}
