﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2014 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.EvolutionTracking {
  [StorableClass]
  [Item("GenealogGraphNode", "A class representing a node in the GenealogyGraph")]
  public class GenealogyGraphNode : Vertex, IGenealogyGraphNode {
    [StorableConstructor]
    protected GenealogyGraphNode(bool deserializing) : base(deserializing) { }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new GenealogyGraphNode(this, cloner);
    }
    protected GenealogyGraphNode(GenealogyGraphNode original, Cloner cloner)
      : base(original, cloner) {
    }

    public GenealogyGraphNode() { }

    public new List<IGenealogyGraphArc> InArcs {
      get {
        return base.InArcs.Cast<IGenealogyGraphArc>().ToList();
      }
    }
    public new List<IGenealogyGraphArc> OutArcs {
      get {
        return base.OutArcs.Cast<IGenealogyGraphArc>().ToList();
      }
    }
    public IEnumerable<IGenealogyGraphNode> Ancestors() {
      // for performance, we use a hashset for lookup and a list for iteration
      var nodes = new HashSet<IGenealogyGraphNode> { this };
      var list = new List<IGenealogyGraphNode> { this };
      int i = 0;
      while (i != list.Count) {
        if (list[i].InArcs != null) {
          foreach (var e in list[i].InArcs) {
            if (nodes.Contains(e.Source)) continue;
            nodes.Add(e.Source);
            list.Add(e.Source);
          }
        }
        ++i;
      }
      return list;
    }
    /// <summary>
    /// Performs a downward-breath-traversal of the genealogy graph using the nodes OutArcs
    /// </summary>
    /// <returns>All the descendants of the current node</returns>
    public IEnumerable<IGenealogyGraphNode> Descendants() {
      var nodes = new HashSet<IGenealogyGraphNode> { this };
      var list = new List<IGenealogyGraphNode> { this };
      int i = 0;
      while (i != list.Count) {
        if (list[i].OutArcs != null) {
          foreach (var e in list[i].OutArcs) {
            if (nodes.Contains(e.Target)) continue;
            nodes.Add(e.Target);
            list.Add(e.Target);
          }
        }
        ++i;
      }
      return list;
    }
    [Storable]
    private double rank;
    public double Rank {
      get { return rank; }
      set { rank = value; }
    }
    [Storable]
    private double quality;
    public double Quality {
      get { return quality; }
      set { quality = value; }
    }
    [Storable]
    private bool isElite;
    public bool IsElite {
      get { return isElite; }
      set { isElite = value; }
    }
    public int CompareTo(IGenealogyGraphNode other) {
      return Quality.CompareTo(other.Quality);
    }
  }

  [StorableClass]
  [Item("GenealogyGraphNode", "A genealogy graph node which also has a Content")]
  public class GenealogyGraphNode<T> : GenealogyGraphNode, IGenealogyGraphNode<T> {
    public T Content { get; set; }
  }
}
