﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2014 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.EvolutionTracking {
  [StorableClass]
  [Item("AfterCrossoverOperator", "A generic operator that can record genealogical relationships between crossover parents and children.")]
  public class AfterCrossoverOperator<TGraph, TVertex, TContent> : EvolutionTrackingOperator<TGraph, TVertex, TContent>
    where TGraph : class,IGenealogyGraph<TVertex, TContent>
    where TVertex : class,IGenealogyGraphNode<TContent>, new()
    where TContent : class,IItem {

    private string parentsParameterName;
    public string ParentsParameterName {
      get { return parentsParameterName; }
      set {
        if (parentsParameterName != null && Parameters.ContainsKey(parentsParameterName)) Parameters.Remove(parentsParameterName); // remove old parameter
        parentsParameterName = value;
        if (!Parameters.ContainsKey(parentsParameterName)) Parameters.Add(new ScopeTreeLookupParameter<TContent>(parentsParameterName));
      }
    }
    private string childParameterName;
    public string ChildParameterName {
      get { return childParameterName; }
      set {
        if (childParameterName != null && Parameters.ContainsKey(childParameterName)) Parameters.Remove(childParameterName);
        childParameterName = value;
        if (!Parameters.ContainsKey(childParameterName)) Parameters.Add(new LookupParameter<TContent>(childParameterName));
      }
    }

    public IScopeTreeLookupParameter<TContent> ParentsParameter {
      get { return (IScopeTreeLookupParameter<TContent>)Parameters[ParentsParameterName]; }
    }

    public ILookupParameter<TContent> ChildParameter {
      get { return (ILookupParameter<TContent>)Parameters[ChildParameterName]; }
    }

    protected AfterCrossoverOperator(AfterCrossoverOperator<TGraph, TVertex, TContent> original, Cloner cloner)
      : base(original, cloner) {
      ParentsParameterName = original.ParentsParameterName;
      ChildParameterName = original.ChildParameterName;
    }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new AfterCrossoverOperator<TGraph, TVertex, TContent>(this, cloner);
    }

    public AfterCrossoverOperator() { }

    public override IOperation Apply() {
      //      var parentVertices = ParentsParameter.ActualValue.Select(x => GenealogyGraph[x].Last()).ToList();
      var lastGen = GenealogyGraph.Ranks[Generations.Value].ToList();
      var parentVertices = ExecutionContext.Scope.SubScopes.Select(s => lastGen[int.Parse(s.Name)]).ToList();
      var childVertex = new TVertex {
        Content = ChildParameter.ActualValue,
        Rank = parentVertices[0].Rank + 1
      };
      GenealogyGraph.AddVertex(childVertex);
      foreach (var v in parentVertices) {
        childVertex.AddReverseArc(v);
        v.AddForwardArc(childVertex);
      }
      return base.Apply();
    }
  }
}
